import six.moves
import base64
import datetime
import hashlib
import re


def _ruby_b64encode(value):
    """The Ruby function Base64.encode64 automatically breaks things up
    into 60-character chunks.
    """
    b64 = base64.b64encode(value)
    for i in six.moves.range(0, len(b64), 60):
        yield b64[i : i + 60].decode()


def ruby_b64encode(value):
    return "\n".join(_ruby_b64encode(value))


def sha1_base64(value):
    """An implementation of Mixlib::Authentication::Digester."""
    return ruby_b64encode(hashlib.sha1(value.encode()).digest())


class UTC(datetime.tzinfo):
    """UTC timezone stub."""

    ZERO = datetime.timedelta(0)

    def utcoffset(self, dt):
        return self.ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return self.ZERO


utc = UTC()


def canonical_time(timestamp):
    if timestamp.tzinfo is not None:
        timestamp = timestamp.astimezone(utc).replace(tzinfo=None)
    return timestamp.replace(microsecond=0).isoformat() + "Z"


canonical_path_regex = re.compile(r"/+")


def canonical_path(path):
    path = canonical_path_regex.sub("/", path)
    if len(path) > 1:
        path = path.rstrip("/")
    return path


def canonical_request(http_method, path, hashed_body, timestamp, user_id):
    # Canonicalize request parameters
    http_method = http_method.upper()
    path = canonical_path(path)
    if isinstance(timestamp, datetime.datetime):
        timestamp = canonical_time(timestamp)
    hashed_path = sha1_base64(path)
    return (
        "Method:%(http_method)s\n"
        "Hashed Path:%(hashed_path)s\n"
        "X-Ops-Content-Hash:%(hashed_body)s\n"
        "X-Ops-Timestamp:%(timestamp)s\n"
        "X-Ops-UserId:%(user_id)s" % vars()
    )


def sign_request(key, http_method, path, body, host, timestamp, user_id):
    """Generate the needed headers for the Opscode authentication protocol."""
    timestamp = canonical_time(timestamp)
    hashed_body = sha1_base64(body or "")

    # Simple headers
    headers = {
        "x-ops-sign": "version=1.0",
        "x-ops-userid": user_id,
        "x-ops-timestamp": timestamp,
        "x-ops-content-hash": hashed_body,
    }

    # Create RSA signature
    req = canonical_request(http_method, path, hashed_body, timestamp, user_id)
    sig = _ruby_b64encode(key.private_encrypt(req))
    for i, line in enumerate(sig):
        headers["x-ops-authorization-%s" % (i + 1)] = line
    return headers



import os

import unittest

from chef.rsa import Key, SSLError
from chef.tests import TEST_ROOT, skipSlowTest


class RSATestCase(unittest.TestCase):
    def test_load_private(self):
        key = Key(os.path.join(TEST_ROOT, "client.pem"))
        self.assertFalse(key.public)

    def test_load_public(self):
        key = Key(os.path.join(TEST_ROOT, "client_pub.pem"))
        self.assertTrue(key.public)

    def test_private_export(self):
        key = Key(os.path.join(TEST_ROOT, "client.pem"))
        raw = open(os.path.join(TEST_ROOT, "client.pem"), "rb").read()
        self.assertTrue(key.private_export().strip(), raw.strip())

    def test_public_export(self):
        key = Key(os.path.join(TEST_ROOT, "client.pem"))
        raw = open(os.path.join(TEST_ROOT, "client_pub.pem"), "rb").read()
        self.assertTrue(key.public_export().strip(), raw.strip())

    def test_private_export_pubkey(self):
        key = Key(os.path.join(TEST_ROOT, "client_pub.pem"))
        with self.assertRaises(SSLError):
            key.private_export()

    def test_public_export_pubkey(self):
        key = Key(os.path.join(TEST_ROOT, "client_pub.pem"))
        raw = open(os.path.join(TEST_ROOT, "client_pub.pem"), "rb").read()
        self.assertTrue(key.public_export().strip(), raw.strip())

    def test_encrypt_decrypt(self):
        key = Key(os.path.join(TEST_ROOT, "client.pem"))
        msg = "Test string!"
        self.assertEqual(key.public_decrypt(key.private_encrypt(msg)), msg)

    def test_encrypt_decrypt_pubkey(self):
        key = Key(os.path.join(TEST_ROOT, "client.pem"))
        pubkey = Key(os.path.join(TEST_ROOT, "client_pub.pem"))
        msg = "Test string!"
        self.assertEqual(pubkey.public_decrypt(key.private_encrypt(msg)), msg)

    def test_generate(self):
        key = Key.generate()
        msg = "Test string!"
        self.assertEqual(key.public_decrypt(key.private_encrypt(msg)), msg)

    def test_generate_load(self):
        key = Key.generate()
        key2 = Key(key.private_export())
        self.assertFalse(key2.public)
        key3 = Key(key.public_export())
        self.assertTrue(key3.public)

    def test_load_pem_string(self):
        key = Key(open(os.path.join(TEST_ROOT, "client.pem"), "rb").read())
        self.assertFalse(key.public)

    def test_load_public_pem_string(self):
        key = Key(open(os.path.join(TEST_ROOT, "client_pub.pem"), "rb").read())
        self.assertTrue(key.public)



from unittest import mock

from chef.fabric import chef_roledefs
from chef.tests import ChefTestCase, mockSearch


class FabricTestCase(ChefTestCase):
    @mock.patch("chef.search.Search")
    def test_roledef(self, MockSearch):
        search_data = {
            ("role", "*:*"): {},
        }
        search_mock_memo = {}

        def search_mock(index, q="*:*", *args, **kwargs):
            data = search_data[index, q]
            search_mock_inst = search_mock_memo.get((index, q))
            if search_mock_inst is None:
                search_mock_inst = search_mock_memo[index, q] = mock.Mock()
                search_mock_inst.data = data
            return search_mock_inst

        MockSearch.side_effect = search_mock
        print(MockSearch("role").data)

    @mockSearch({("role", "*:*"): {1: 2}})
    def test_roledef2(self, MockSearch):
        print(MockSearch("role").data)



import six
from collections.abc import Sequence
import copy
import six.moves.urllib.parse

from chef.api import ChefAPI
from chef.base import ChefQuery, ChefObject


class SearchRow(dict):
    """A single row in a search result."""

    def __init__(self, row, api):
        super(SearchRow, self).__init__(row)
        self.api = api
        self._object = None

    @property
    def object(self):
        if self._object is None:
            # Decode Chef class name
            chef_class = self.get("json_class", "")
            if chef_class.startswith("Chef::"):
                chef_class = chef_class[6:]
            if chef_class == "ApiClient":
                chef_class = "Client"  # Special case since I don't match the Ruby name.
            cls = ChefObject.types.get(chef_class.lower())
            if not cls:
                raise ValueError("Unknown class %s" % chef_class)
            self._object = cls.from_search(self, api=self.api)
        return self._object


class Search(Sequence):
    """A search of the Chef index.

    The only required argument is the index name to search (eg. node, role, etc).
    The second, optional argument can be any Solr search query, with the same semantics
    as Chef.

    Example::

        for row in Search('node', 'roles:app'):
            print row['roles']
            print row.object.name

    .. versionadded:: 0.1
    """

    url = "/search"

    def __init__(self, index, q="*:*", rows=1000, start=0, api=None):
        self.name = index
        self.api = api or ChefAPI.get_global()
        self._args = dict(q=q, rows=rows, start=start)
        self.url = (
            self.__class__.url
            + "/"
            + self.name
            + "?"
            + six.moves.urllib.parse.urlencode(self._args)
        )

    @property
    def data(self):
        if not hasattr(self, "_data"):
            self._data = self.api[self.url]
        return self._data

    @property
    def total(self):
        return self.data["total"]

    def query(self, query):
        args = copy.copy(self._args)
        args["q"] = query
        return self.__class__(self.name, api=self.api, **args)

    def rows(self, rows):
        args = copy.copy(self._args)
        args["rows"] = rows
        return self.__class__(self.name, api=self.api, **args)

    def start(self, start):
        args = copy.copy(self._args)
        args["start"] = start
        return self.__class__(self.name, api=self.api, **args)

    def __len__(self):
        return len(self.data["rows"])

    def __getitem__(self, value):
        if isinstance(value, slice):
            if value.step is not None and value.step != 1:
                raise ValueError("Cannot use a step other than 1")
            return self.start(self._args["start"] + value.start).rows(
                value.stop - value.start
            )
        if isinstance(value, six.string_types):
            return self[self.index(value)]
        row_value = self.data["rows"][value]
        # Check for null rows, just in case
        if row_value is None:
            return None
        return SearchRow(row_value, self.api)

    def __contains__(self, name):
        for row in self:
            if row.object.name == name:
                return True
        return False

    def index(self, name):
        for i, row in enumerate(self):
            if row.object.name == name:
                return i
        raise ValueError("%s not in search" % name)

    def __call__(self, query):
        return self.query(query)

    @classmethod
    def list(cls, api=None):
        api = api or ChefAPI.get_global()
        names = [name for name, url in six.iteritems(api[cls.url])]
        return ChefQuery(cls, names, api)



import datetime
import logging
import os
import re
import json
import socket
import subprocess
import threading
import weakref
import six
import subprocess
from pathlib import Path

import pkg_resources

import requests

from chef.auth import sign_request
from chef.exceptions import ChefServerError
from chef.rsa import Key
import chef.utils.json
from chef.utils.file import walk_backwards
import chef.bin

api_stack = threading.local()
log = logging.getLogger("chef.api")

config_ruby_script = """
require 'chef'
Chef::Config.from_file('%s')
puts Chef::Config.configuration.to_json
""".strip()


def api_stack_value():
    if not hasattr(api_stack, "value"):
        api_stack.value = []
    return api_stack.value


class UnknownRubyExpression(Exception):
    """Token exception for unprocessed Ruby expressions."""


class ChefAPI(object):
    """The ChefAPI object is a wrapper for a single Chef server.

    .. admonition:: The API stack

        PyChef maintains a stack of :class:`ChefAPI` objects to be use with
        other methods if an API object isn't given explicitly. The first
        ChefAPI created will become the default, though you can set a specific
        default using :meth:`ChefAPI.set_default`. You can also use a ChefAPI
        as a context manager to create a scoped default::

            with ChefAPI('http://localhost:4000', 'client.pem', 'admin'):
                n = Node('web1')
    """

    ruby_value_re = re.compile(r"#\{([^}]+)\}")
    env_value_re = re.compile(r"ENV\[(.+)\]")
    ruby_string_re = re.compile(r'^\s*(["\'])(.*?)\1\s*$')

    def __init__(self, url, key, client, version="0.10.8", headers={}, ssl_verify=True):
        self.url = url.rstrip("/")
        self.parsed_url = six.moves.urllib.parse.urlparse(self.url)
        if not isinstance(key, Key):
            key = Key(key)
        if not key.key:
            raise ValueError("ChefAPI attribute 'key' was invalid.")
        self.key = key
        self.client = client
        self.version = version
        self.headers = dict((k.lower(), v) for k, v in six.iteritems(headers))
        self.version_parsed = pkg_resources.parse_version(self.version)
        self.platform = self.parsed_url.hostname == "api.opscode.com"
        self.ssl_verify = ssl_verify
        if not api_stack_value():
            self.set_default()

    @classmethod
    def from_config_file(cls, path):
        """Load Chef API paraters from a config file. Returns None if the
        config can't be used.
        """
        log.debug('Trying to load from "%s"', path)
        if not os.path.isfile(path) or not os.access(path, os.R_OK):
            # Can't even read the config file
            log.debug('Unable to read config file "%s"', path)
            return
        url = key_path = client_name = None
        ssl_verify = True
        for line in open(path):
            if not line.strip() or line.startswith("#"):
                continue  # Skip blanks and comments
            parts = line.split(None, 1)
            if len(parts) != 2:
                continue  # Not a simple key/value, we can't parse it anyway
            key, value = parts
            md = cls.ruby_string_re.search(value)
            if md:
                value = md.group(2)
            elif key == "ssl_verify_mode":
                log.debug("Found ssl_verify_mode: %r", value)
                ssl_verify = value.strip() != ":verify_none"
                log.debug("ssl_verify = %s", ssl_verify)
            else:
                # Not a string, don't even try
                log.debug(
                    "Value for {0} does not look like a string: {1}".format(key, value)
                )
                continue

            def _ruby_value(match):
                expr = match.group(1).strip()
                if expr == "current_dir":
                    return os.path.dirname(path)
                envmatch = cls.env_value_re.match(expr)
                if envmatch:
                    envmatch = envmatch.group(1).strip('"').strip("'")
                    return os.environ.get(envmatch) or ""
                log.debug('Unknown ruby expression in line "%s"', line)
                raise UnknownRubyExpression

            try:
                value = cls.ruby_value_re.sub(_ruby_value, value)
            except UnknownRubyExpression:
                continue
            if key == "chef_server_url":
                log.debug("Found URL: %r", value)
                url = value
            elif key == "node_name":
                log.debug("Found client name: %r", value)
                client_name = value
            elif key == "client_key":
                log.debug("Found key path: %r", value)
                key_path = value
                if not os.path.isabs(key_path):
                    # Relative paths are relative to the config file
                    key_path = os.path.abspath(
                        os.path.join(os.path.dirname(path), key_path)
                    )

        if not (url and client_name and key_path):
            # No URL, no chance this was valid, try running Ruby
            log.debug("No Chef server config found, trying Ruby parse")
            url = key_path = client_name = None
            proc = subprocess.Popen(
                "ruby", stdin=subprocess.PIPE, stdout=subprocess.PIPE
            )
            script = config_ruby_script % path.replace("\\", "\\\\").replace("'", "\\'")
            out, err = proc.communicate(script.encode())
            if proc.returncode == 0 and out.strip():
                data = chef.utils.json.loads(out.decode())
                log.debug("Ruby parse succeeded with %r", data)
                url = data.get("chef_server_url")
                client_name = data.get("node_name")
                key_path = data.get("client_key")
                if key_path and not os.path.isabs(key_path):
                    # Relative paths are relative to the config file
                    key_path = os.path.abspath(
                        os.path.join(os.path.dirname(path), key_path)
                    )
            else:
                log.debug(
                    "Ruby parse failed with exit code %s: %s",
                    proc.returncode,
                    out.strip(),
                )
        if not url:
            # Still no URL, can't use this config
            log.debug("Still no Chef server URL found")
            return
        if not key_path:
            # Try and use ./client.pem
            key_path = os.path.join(os.path.dirname(path), "client.pem")
        if not os.path.isfile(key_path) or not os.access(key_path, os.R_OK):
            # Can't read the client key
            log.debug('Unable to read key file "%s"', key_path)
            return
        if not client_name:
            client_name = socket.getfqdn()
        return cls(url, key_path, client_name, ssl_verify=ssl_verify)

    @staticmethod
    def get_global():
        """Return the API on the top of the stack."""
        while api_stack_value():
            api = api_stack_value()[-1]()
            if api is not None:
                return api
            del api_stack_value()[-1]

    def set_default(self):
        """Make this the default API in the stack. Returns the old default if any."""
        old = None
        if api_stack_value():
            old = api_stack_value().pop(0)
        api_stack_value().insert(0, weakref.ref(self))
        return old

    def __enter__(self):
        api_stack_value().append(weakref.ref(self))
        return self

    def __exit__(self, type, value, traceback):
        del api_stack_value()[-1]

    def _request(self, method, url, data, headers):
        return requests.api.request(
            method, url, headers=headers, data=data, verify=self.ssl_verify
        )

    def request(self, method, path, headers={}, data=None):
        auth_headers = sign_request(
            key=self.key,
            http_method=method,
            path=self.parsed_url.path + path.split("?", 1)[0],
            body=data,
            host=self.parsed_url.netloc,
            timestamp=datetime.datetime.utcnow(),
            user_id=self.client,
        )
        request_headers = {}
        request_headers.update(self.headers)
        request_headers.update(dict((k.lower(), v) for k, v in six.iteritems(headers)))
        request_headers["x-chef-version"] = self.version
        request_headers.update(auth_headers)
        try:
            response = self._request(
                method,
                self.url + path,
                data,
                dict((k.capitalize(), v) for k, v in six.iteritems(request_headers)),
            )
        except requests.ConnectionError as e:
            raise ChefServerError(e.message)

        if not response.ok:
            raise ChefServerError.from_error(response.reason, code=response.status_code)

        return response

    def api_request(self, method, path, headers={}, data=None):
        headers = dict((k.lower(), v) for k, v in six.iteritems(headers))
        headers["accept"] = "application/json"
        if data is not None:
            headers["content-type"] = "application/json"
            data = chef.utils.json.dumps(data)
        response = self.request(method, path, headers, data)
        return response.json()

    def __getitem__(self, path):
        return self.api_request("GET", path)


def autoconfigure_v2():
    """
    Re-uses Chef's configuration parsing & merging logic instead of reimplementing in Python :)
    See load_config_chef.utils.json.rb script for additional info

    Adds support for multiple configuration files from config.d directory.
    All files that end in .rb in the .d directory are loaded; other non-.rb files are ignored.

    For example:
    /etc/chef/config.d
    ~/.chef/config.d/company_settings.rb
    ~/.chef/config.d/ec2_configuration.rb
    ~/.chef/config.d/old_settings.rb.bak

    The old_settings.rb.bak file is ignored because it’s not a configuration file.
    The config.rb, company_settings.rb, and ec2_configuration files are merged together as if they are a single configuration file.

    https://docs.chef.io/workstation/config_rb/#d-directories
    """
    script = os.path.join(
        os.path.dirname(chef.bin.__file__), "load_config_chef.utils.json.rb"
    )
    stdout = subprocess.check_call(script)
    config = json.loads(stdout)

    chef_server_url = config.get("chef_server_url")
    key = config.get("client_key")
    ssl_verify_mode = config.get("ssl_verify_mode")
    node_name = config.get("node_name")

    ssl_verify = ssl_verify_mode != "verify_none"
    return ChefAPI(
        url=chef_server_url, key=key, client=node_name, ssl_verify=ssl_verify
    )


def autoconfigure(base_path=None):
    """Try to find a knife or chef-client config file to load parameters from,
    starting from either the given base path or the current working directory.

    The lookup order mirrors the one from Chef, first all folders from the base
    path are walked back looking for .chef/knife.rb, then ~/.chef/knife.rb,
    and finally /etc/chef/client.rb.

    The first file that is found and can be loaded successfully will be loaded
    into a :class:`ChefAPI` object.
    """
    log.warning(
        "DEPRECATION NOTICE: chef.api.autoconfigure will be removed in a future version of Py3Chef. Please use chef.api.autoconfigure_v2 to fix this warning."
    )
    base_path = base_path or os.getcwd()
    # Scan up the tree for a knife.rb or client.rb. If that fails try looking
    # in /etc/chef. The /etc/chef check will never work in Win32, but it doesn't
    # hurt either.
    for path in walk_backwards(base_path):
        config_path = os.path.join(path, ".chef", "knife.rb")
        api = ChefAPI.from_config_file(config_path)
        if api is not None:
            return api

    # The walk didn't work, try ~/.chef/knife.rb
    config_path = os.path.expanduser(os.path.join("~", ".chef", "knife.rb"))
    api = ChefAPI.from_config_file(config_path)
    if api is not None:
        return api

    # Nothing in the home dir, try /etc/chef/client.rb
    config_path = os.path.join(os.path.sep, "etc", "chef", "client.rb")
    api = ChefAPI.from_config_file(config_path)
    if api is not None:
        return api



from chef import Environment
from chef.exceptions import ChefAPIVersionError
from chef.tests import ChefTestCase, test_chef_api


class EnvironmentTestCase(ChefTestCase):
    def test_version_error_list(self):
        with test_chef_api(version="0.9.0"):
            with self.assertRaises(ChefAPIVersionError):
                Environment.list()

    def test_version_error_create(self):
        with test_chef_api(version="0.9.0"):
            with self.assertRaises(ChefAPIVersionError):
                Environment.create(self.random())

    def test_version_error_init(self):
        with test_chef_api(version="0.9.0"):
            with self.assertRaises(ChefAPIVersionError):
                Environment(self.random())
